import axios from 'axios';

/**
 * Handle authentication and send/receive with the backbone
 */
class Adapter {
  constructor(protocol, config) {
    this.protocol = protocol;
    this.config = config;
    this.axios = axios;
    this.pollTimer = null;
  }

  /**
   * Start listening for inbound messages 
   */
  start() {
    switch (this.getMode()) {
      case 'socket': this.openSocket(); break;
      case 'http': {
        const interval = ('pollInterval' in this.config) ? this.config.pollInterval : 5;
        this.pollEveryXSeconds(interval);
      } break;
    }
  }

  /**
   * Test parsing the message based on the provided protocol schema
   *
   * The message must be successfully json decoded into an object
   * prior to validation
   *
   * At present this returns the validation result which is an
   * object containing a boolean valid field as well as details
   * of any errors.
   * @param {object} message
   * @returns {object}
   */
  validate(message) {
    return this.protocol.validate(message);
  }

  /**
   * Ensure the token in this.credentials is still valid
   * @returns {boolean}
   */
  tokenValid() {
    let valid = false;
    if (this.credentials) {
      const now = new Date().toISOString();
      valid = this.credentials.expiry > now;
    }
    return valid;
  }

  /**
   * Return an http headers object containing a bearer token authorisation header
   * @returns {object}
   */
  getAuthorizationHeader() {
    if (!this.tokenValid())
      return this.auth().then((response) => {
        return {
          Authorization: `Bearer ${response.data.token}`,
        };
      });
    else {
      return Promise.resolve({
        Authorization: `Bearer ${this.credentials.token}`,
      });
    }
  }

  /**
   * Do a client credentials grant and store in this.credentials
   * @returns {object}
   */
  auth() {
    let adapterConfig = this.config;
    return this.axios
      .get(`${adapterConfig.api}/token`, {
        params: {
          client_id: adapterConfig.client_id,
          secret: adapterConfig.secret,
        },
      })
      .then((response) => {
        this.credentials = response.data;
        return response;
      })
      .catch((error) => {
        return Promise.reject(error);
      });
  }

  /**
   * Setup a timer to call poll every interval seconds
   * @param {int} interval 
   */
  pollEveryXSeconds(interval) {
    if (this.pollTimer) clearTimeout(this.pollTimer);
    this.pollTimer = setTimeout(this.pollEveryXSeconds.bind(this, interval), interval*1000);
    this.poll();
  }

  /**
   * Call the GET /receive endpoint and process the messages with decode
   *
   * Returns the response
   * @param {boolean} is_retry
   * @returns {object}
   */
  poll(is_retry = false) {
    let adapterConfig = this.config;
    return this.getAuthorizationHeader()
      .then((headers) => {
        return this.axios.get(`${adapterConfig.api}/receive`, {
          headers,
        });
      })
      .then((response) => {
        response.data.forEach((message) => {
          this.receive(message); 
        });
        return response;
      })
      .catch((error) => {
        let retry = false;
        switch (error.response.status) {
          case 403:
            {
              this.credentials = null;
              retry = true;
            }
            break;
          case 503: {
            retry = true;
          }
        }
        if (retry && !is_retry) return this.poll(true);
        else return Promise.reject(error);
      });
  }

  /**
   * Process a message received either from poll 
   * or from socket.onMessage 
   * @param {object} message 
   */
  receive(message) {
    try {
      const parsed = JSON.parse(message.message);
      const validation = this.validate(parsed);
      if (validation.valid) {
        const type = this.protocol.getType(parsed);
        this.protocol.decode(type, parsed);
      } else {
        this.protocol.receivedInvalid(parsed, validation);
      }
    } catch(err) {
      console.error('receive failed', err);
    }
  }

  /**
   * Decide whether to use http or socket for comms 
   * with the backbone 
   * @returns string 
   */
  getMode() {
    return ('mode' in this.config) ? this.config.mode : 'http'; 
  }

  /**
   * Create a new socket connection and send an authentication
   * message. Websocket doesn't support headers so as soon 
   * as the socket is open a message is sent containing 
   * the authorisation header to authenticate the connection 
   */
  openSocket() {
    this.getAuthorizationHeader()
    .then((headers) => {

      const protocols = [];
      
      const socket = new WebSocket(
        this.config.socket, 
        protocols, 
      );

      socket.addEventListener('open', () => {
        socket.send(JSON.stringify({ action: 'auth', headers }));
      });

      socket.addEventListener('message', (event) => {
        console.log('message received');
        try {
          const message = JSON.parse(event.data);
          this.receive(message);
        } catch (err) {
          console.error('invalid message', typeof event.data);
        }
      });

      socket.addEventListener('close', () => {
        this.closeSocket();
      });

      this.socket = socket;
      return socket; 
    });
  }
  
  /**
   * If the socket connection is lost ensure that a new socket 
   * is opened rather than continuing to send messages on this
   * socket 
   */
  closeSocket() {
    this.socket = null;
  }

  /**
   * Publish a message to the backbone with the specified topic
   *
   * Messages should be passed through encode before sending
   * @param {string} topic
   * @param {string} body
   * @returns
   */
  publish(topic, body) {
    let mode = this.getMode();
    let response;
    switch (mode) {
      case 'http': response = this.http_publish(topic, body); break;
      case 'socket': response = this.socket_publish(topic, body); break;
    }
    return response;
  }

  /**
   * Send via a POST to /send
   * @param {string} topic 
   * @param {string} body 
   * @param {boolean} is_retry
   * @returns Promise
   */
  http_publish(topic, body, is_retry = false) {
    let adapterConfig = this.config;
    return this.getAuthorizationHeader()
      .then((headers) => {
        return this.axios.post(
          `${adapterConfig.api}/send`,
          {
            topic,
            body,
          },
          {
            headers,
          }
        );
      })
      .then((response) => {
        return response;
      })
      .catch((error) => {
        let retry = false;
        switch (error.response.status) {
          case 403:
            {
              this.credentials = null;
              retry = true;
            }
            break;
          case 503: {
            retry = true;
          }
        }
        if (retry && !is_retry) return this.http_publish(topic, body, true);
        else return Promise.reject(error);
      });
  }

  /**
   * Publish via websocket socket.send
   * @param {string} topic 
   * @param {string} body 
   */
  socket_publish(topic, body) {
    if (this.socket) {
      this.socket.send(JSON.stringify({ topic, message: body }));
    } else {
      this.openSocket().then((socket) => {
        socket.send(JSON.stringify({ topic, message: body }));
      });
    }
  }

  /**
   * Broadcast the message on the backbone
   *
   * Broadcast messages bypass the normal publisher queues
   * this means they can be used to deliver messages more
   * quickly in an emergency scenario.
   *
   * Messages should be passed through encode before sending
   * @param {string} body
   * @returns
   */
  broadcast(body) {
    let mode = this.getMode();
    let response;
    switch (mode) {
      case 'http': response = this.http_broadcast(body); break;
      case 'socket': response = this.socket_broadcast(body); break;
    }
    return response;
  }

  /**
   * Broadcast via POST to /notify
   * @param {string} body 
   * @param {boolean} is_retry
   * @returns Promise
   */
  http_broadcast(body, is_retry = false) {
    let adapterConfig = this.config;
    return this.getAuthorizationHeader()
      .then((headers) => {
        return this.axios.post(
          `${adapterConfig.api}/notify`,
          {
            body,
          },
          {
            headers,
          }
        );
      })
      .then((response) => {
        return response;
      })
      .catch((error) => {
        let retry = false;
        switch (error.response.status) {
          case 403:
            {
              this.credentials = null;
              retry = true;
            }
            break;
          case 503: {
            retry = true;
          }
        }
        if (retry && !is_retry) return this.http_broadcast(body, true);
        else return Promise.reject(error);
      });
  }

  /**
   * Broadcast via websocket socket.send
   * @param {string} body 
   */
  socket_broadcast(body) {
    if (this.socket) {
      this.socket.send(JSON.stringify({ message: body }));
    } else {
      this.openSocket().then((socket) => {
        socket.send(JSON.stringify({ message: body }));
      });
    }
  }
}

export { Adapter };
